
static int librad_dodns = 0;
static int librad_max_attributes = 0;

char librad_errstr[1024];

static DICT_ATTR	*base_attributes[256];
static DICT_ATTR	*dictionary_attributes = NULL;
static DICT_VALUE	*dictionary_values = NULL;
static DICT_VENDOR	*dictionary_vendors = NULL;
static int saltoffset = 0; /* used to make all salts in a packet unique */

typedef struct {
	const char *str;
	int token;
} TOKEN;

static const TOKEN tokens[] = {
	{ "=~", T_OP_REG_EQ,	}, /* order is important! */
	{ "!~", T_OP_REG_NE,	},
	{ "{",	T_LCBRACE,	},
	{ "}",	T_RCBRACE,	},
	{ "(",	T_LBRACE,	},
	{ ")",	T_RBRACE,	},
	{ ",",	T_COMMA,	},
	{ "+=",	T_OP_ADD,	},
	{ "-=",	T_OP_SUB,	},
	{ ":=",	T_OP_SET,	},
	{ "=*", T_OP_CMP_TRUE,  },
	{ "!*", T_OP_CMP_FALSE, },
	{ "==",	T_OP_CMP_EQ,	},
	{ "=",	T_OP_EQ,	},
	{ "!=",	T_OP_NE,	},
	{ ">=",	T_OP_GE,	},
	{ ">",	T_OP_GT,	},
	{ "<=",	T_OP_LE,	},
	{ "<",	T_OP_LT,	},
	{ "#",	T_HASH,		},
	{ ";",	T_SEMICOLON,	},
	{ NULL, 0,		},
};

#define TOKEN_MATCH(bptr, tptr) ( (tptr)[0] == (bptr)[0] && ((tptr)[1] == (bptr)[1] || (tptr)[1] == 0))

static const char *months[] = {
        "jan", "feb", "mar", "apr", "may", "jun",
        "jul", "aug", "sep", "oct", "nov", "dec" };

static uint8_t random_vector_pool[AUTH_VECTOR_LEN*2];

static const char *packet_codes[] = {
  "",
  "Access-Request",
  "Access-Accept",
  "Access-Reject",
  "Accounting-Request",
  "Accounting-Response",
  "Accounting-Status",
  "Password-Request",
  "Password-Accept",
  "Password-Reject",
  "Accounting-Message",
  "Access-Challenge",
  "Status-Server",
  "Status-Client"
};

/*
 *	Several handy miscellaneous functions.
 */
char *	ip_hostname (char *buf, size_t buflen, uint32_t ipaddr);
uint32_t	ip_getaddr (const char *);
char *		ip_ntoa(char *, uint32_t);
uint32_t	ip_addr(const char *);
//char		*strNcpy(char *dest, const char *src, int n);
//void		rad_lowercase(char *str);
//void		rad_rmspace(char *str);
//int		rad_lockfd(int fd, int lock_len);
//int		rad_lockfd_nonblock(int fd, int lock_len);
//int		rad_unlockfd(int fd, int lock_len);

LRAD_TOKEN	gettoken(char **ptr, char *buf, int buflen);

char *ip_ntoa(char *buffer, uint32_t ipaddr)
{
	ipaddr = ntohl(ipaddr);

	sprintf(buffer, "%d.%d.%d.%d",
		(ipaddr >> 24) & 0xff,
		(ipaddr >> 16) & 0xff,
		(ipaddr >>  8) & 0xff,
		(ipaddr      ) & 0xff);
	return buffer;
}

char * ip_hostname(char *buf, size_t buflen, uint32_t ipaddr)
{
	struct		hostent *hp;
#ifdef GETHOSTBYADDRRSTYLE
#if (GETHOSTBYADDRRSTYLE == SYSVSTYLE) || (GETHOSTBYADDRRSTYLE == GNUSTYLE)
	char buffer[2048];
	struct hostent result;
	int error;
#endif
#endif

	/*
	 *	No DNS: don't look up host names
	 */
	if (librad_dodns == 0) {
		ip_ntoa(buf, ipaddr);
		return buf;
	}

#ifdef GETHOSTBYADDRRSTYLE
#if GETHOSTBYADDRRSTYLE == SYSVSTYLE
	hp = gethostbyaddr_r((char *)&ipaddr, sizeof(struct in_addr), AF_INET, &result, buffer, sizeof(buffer), &error);
#elif GETHOSTBYADDRRSTYLE == GNUSTYLE
	gethostbyaddr_r((char *)&ipaddr, sizeof(struct in_addr), AF_INET, &result, buffer, sizeof(buffer), &hp, &error);
#else
	hp = gethostbyaddr((char *)&ipaddr, sizeof(struct in_addr), AF_INET);
#endif
#else
	hp = gethostbyaddr((char *)&ipaddr, sizeof(struct in_addr), AF_INET);
#endif
	if ((hp == NULL) ||
	    (strlen((char *)hp->h_name) >= buflen)) {
		ip_ntoa(buf, ipaddr);
		return buf;
	}

	strncpy(buf, (char *)hp->h_name, buflen);
	return buf;
}

uint32_t ip_addr(const char *ip_str)
{
	struct in_addr	in;

	if (inet_aton(ip_str, &in) == 0)
		return INADDR_NONE;
	return in.s_addr;
}

uint32_t ip_getaddr(const char *host)
{
	struct hostent	*hp;
	uint32_t	 a;
#ifdef HAVE_GETHOSTBYNAME_R
	struct hostent result;
	int error;
	char buffer[2048];
#endif

	if ((a = ip_addr(host)) != htonl(INADDR_NONE))
		return a;

#ifndef HAVE_GETHOSTBYNAME_R
	hp = gethostbyname(host);
#else
	hp = gethostbyname_r(host, &result, buffer, sizeof(buffer), &error);
#endif
	if (hp == NULL) {
		return htonl((uint32_t) INADDR_NONE);
	}

	/*
	 *	Paranoia from a Bind vulnerability.  An attacker
	 *	can manipulate DNS entries to change the length of the
	 *	address.  If the length isn't 4, something's wrong.
	 */
	if (hp->h_length != sizeof(uint32_t)) {
		return htonl((uint32_t) INADDR_NONE);
	}

	memcpy(&a, hp->h_addr, sizeof(uint32_t));
	return a;
}

static int calc_replydigest(RADIUS_PACKET *packet, RADIUS_PACKET *original, const char *secret)
{
	uint8_t		calc_digest[AUTH_VECTOR_LEN];
	struct MD5Context		context;

	/*
	 *  Copy the original vector in place.
	 */
	memcpy(packet->data + 4, original->vector, sizeof(original->vector));

	/*
	 *  MD5(packet + secret);
	 */
	MD5Init(&context);
	MD5Update(&context, packet->data, packet->data_len);
	MD5Update(&context, secret, strlen(secret));
	MD5Final(calc_digest, &context);

	/*
	 *  Copy the packet's vector back to the packet.
	 */
	memcpy(packet->data + 4, packet->vector, sizeof(packet->vector));

	/*
	 *	Return 0 if OK, 2 if not OK.
	 */
	packet->verified =
		memcmp(packet->vector, calc_digest, sizeof(packet->vector)) ? 2 : 0;
	return packet->verified;
}

static int calc_acctdigest(RADIUS_PACKET *packet, const char *secret)
{
	u_char		digest[AUTH_VECTOR_LEN];
	struct MD5Context	context;

	/*
	 *	Older clients have the authentication vector set to
	 *	all zeros. Return `1' in that case.
	 */
	memset(digest, 0, sizeof(digest));
	if (memcmp(packet->vector, digest, AUTH_VECTOR_LEN) == 0) {
		packet->verified = 1;
		return 1;
	}

	/*
	 *	Zero out the auth_vector in the received packet.
	 *	Then append the shared secret to the received packet,
	 *	and calculate the MD5 sum. This must be the same
	 *	as the original MD5 sum (packet->vector).
	 */
	memset(packet->data + 4, 0, AUTH_VECTOR_LEN);
	
	/*
	 *  MD5(packet + secret);
	 */
	MD5Init(&context);
	MD5Update(&context, packet->data, packet->data_len);
	MD5Update(&context, secret, strlen(secret));
	MD5Final(digest, &context);

	/*
	 *	Return 0 if OK, 2 if not OK.
	 */
	packet->verified =
	memcmp(digest, packet->vector, AUTH_VECTOR_LEN) ? 2 : 0;

	return packet->verified;
}

DICT_VALUE * dict_valbyattr(int attr, int val)
{
	DICT_VALUE	*v;

	for (v = dictionary_values; v; v = v->next) {
		if (v->attr == attr && v->value == val)
			return v;
	}

	return NULL;
}

static char *mystrtok(char **ptr, const char *sep)
{
	char	*res;

	if (**ptr == 0)
		return NULL;
	while (**ptr && strchr(sep, **ptr))
		(*ptr)++;
	if (**ptr == 0)
		return NULL;
	res = *ptr;
	while (**ptr && strchr(sep, **ptr) == NULL)
		(*ptr)++;
	if (**ptr != 0)
		*(*ptr)++ = 0;
	return res;
}

static int gettime(const char *valstr, time_t *lvalue)
{
	int		i;
	time_t		t;
	struct tm	*tm, s_tm;
	char		buf[64];
	char		*p;
	char		*f[3];

	time(&t);
	tm = localtime_r(&t, &s_tm);

	strncpy(buf, valstr, sizeof(buf));

	p = buf;
	f[0] = mystrtok(&p, " \t");
	f[1] = mystrtok(&p, " \t");
	f[2] = mystrtok(&p, " \t");
	if (!f[0] || !f[1] || !f[2]) return -1;

	/*
	 *  The month is text, which allows us to find it easily.
	 */
	tm->tm_mon = 12;
	for (i = 0; i < 3; i++) {
		if (isalpha( (int) *f[i])) {	
			/*
			 *  Bubble the month to the front of the list
			 */
			p = f[0];
			f[0] = f[i];
			f[i] = p;

			for (i = 0; i < 12; i++) {
				if (strncasecmp(months[i], f[0], 3) == 0) {
					tm->tm_mon = i;
					break;
				}
			}
		}
	}

	/* month not found? */
	if (tm->tm_mon == 12) return -1;

	/*
	 *  The year may be in f[1], or in f[2]
	 */
	tm->tm_year = atoi(f[1]);
	tm->tm_mday = atoi(f[2]);

	if (tm->tm_year >= 1900) {
		tm->tm_year -= 1900;

	} else {
		/*
		 *  We can't use 2-digit years any more, they make it
		 *  impossible to tell what's the day, and what's the year.
		 */
		if (tm->tm_mday < 1900) return -1;

		/*
		 *  Swap the year and the day.
		 */
		i = tm->tm_year;
		tm->tm_year = tm->tm_mday - 1900;
		tm->tm_mday = i;
	}

	/*
	 *  If the day is out of range, die.
	 */
	if ((tm->tm_mday < 1) || (tm->tm_mday > 31)) {
		return -1;
	}

	/*
	 *  Returns -1 on error.
	 */
	t = mktime(tm);
	if (t == (time_t) -1) return -1;

	*lvalue = t;

	return 0;
}

void librad_md5_calc(unsigned char *output, unsigned char *input, unsigned int inlen)
{
	struct MD5Context	context;

	MD5Init(&context);
	MD5Update(&context, input, inlen);
	MD5Final(output, &context);
}

void pairadd(VALUE_PAIR **first, VALUE_PAIR *add)
{
	VALUE_PAIR *i;

	if (*first == NULL) {
		*first = add;
		return;
	}
	for(i = *first; i->next; i = i->next)
		;
	i->next = add;
}

static LRAD_TOKEN getthing(char **ptr, char *buf, int buflen, int tok, const TOKEN *tokenlist)
{
	char	*s, *p;
	int	quote;
	int	escape;
	int	x;
	const TOKEN	*t;
	LRAD_TOKEN rcode;

	buf[0] = 0;

	/* Skip whitespace */
	p = *ptr;
	while (*p && isspace((int) *p))
		p++;

	if (*p == 0) {
		*ptr = p;
		return T_EOL;
	}

	/*
	 *	Might be a 1 or 2 character token.
	 */
	if (tok) for (t = tokenlist; t->str; t++) {
		if (TOKEN_MATCH(p, t->str)) {
			strcpy(buf, t->str);
			p += strlen(t->str);
			while (isspace((int) *p))
				p++;
			*ptr = p;
			return t->token;
		}
	}

	/* Read word. */
	quote = 0;
	if ((*p == '"') || (*p == '\'') || (*p == '`')) {
		quote = *p;
		p++;
	}
	s = buf;
	escape = 0;

	while (*p && buflen-- > 0) {
		if (escape) {
			escape = 0;
			switch(*p) {
				case 'r':
					*s++ = '\r';
					break;
				case 'n':
					*s++ = '\n';
					break;
				case 't':
					*s++ = '\t';
					break;
				case '"':
					*s++ = '"';
					break;
				case '\'':
					*s++ = '\'';
					break;
				case '`':
					*s++ = '`';
					break;
				default:
					if (*p >= '0' && *p <= '9' &&
					    sscanf(p, "%3o", &x) == 1) {
						*s++ = x;
						p += 2;
					} else
						*s++ = *p;
					break;
			}
			p++;
			continue;
		}
		if (*p == '\\') {
			p++;
			escape = 1;
			continue;
		}
		if (quote && (*p == quote)) {
			p++;
			break;
		}
		if (!quote) {
			if (isspace((int) *p))
				break;
			if (tok) {
				for (t = tokenlist; t->str; t++)
					if (TOKEN_MATCH(p, t->str))
						break;
				if (t->str != NULL)
					break;
			}
		}
		*s++ = *p++;
	}
	*s++ = 0;

	/* Skip whitespace again. */
	while (*p && isspace((int) *p))
		p++;
	*ptr = p;

	/* we got SOME form of output string, even if it is empty */
	switch (quote) {
		default:
			rcode = T_BARE_WORD;
			break;
		case '\'':
			rcode = T_SINGLE_QUOTED_STRING;
			break;
		case '"':
			rcode = T_DOUBLE_QUOTED_STRING;
			break;
		case '`':
			rcode = T_BACK_QUOTED_STRING;
			break;
	}

	return rcode;
}

LRAD_TOKEN gettoken(char **ptr, char *buf, int buflen)
{
	return getthing(ptr, buf, buflen, 1, tokens);
}

VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
{
	char		*p, *s=0;
	const char	*cp, *cs;
	DICT_VALUE	*dval;

	/*
	 *	Even for integers, dates and ip addresses we
	 *	keep the original string in vp->strvalue.
	 */
	strncpy((char *)vp->strvalue, value, sizeof(vp->strvalue));
	vp->length = strlen(vp->strvalue);

	switch(vp->type) {
		case PW_TYPE_STRING:
			/*
			 *	Already handled above.
			 */
			break;

		case PW_TYPE_IPADDR:
			/*
			 *	FIXME: complain if hostname
			 *	cannot be resolved, or resolve later!
			 */
			if ((p = strrchr(value, '+')) != NULL && !p[1]) {
				cs = s = strdup(value);
				p = strrchr(s, '+');
				*p = 0;
				vp->flags.addport = 1;
			} else {
				p = NULL;
				cs = value;
			}
			vp->lvalue = librad_dodns ? ip_getaddr(cs) : ip_addr(cs);
			if (s) free(s);
			vp->length = 4;
			break;
		case PW_TYPE_INTEGER:
			/*
			 * 	If it starts with a digit, it must
			 * 	be a number (or a range).
			 *
			 *	Note that ALL integers are unsigned!
			 */
			if (isdigit((int) *value)) {
				vp->lvalue = atoi(value);
				vp->length = 4;
			} else if ((dval = dict_valbyname(vp->attribute, value)) == NULL) {
				/*
				 *	Look for the named value for the given
				 *	attribute.
				 */
				ast_log(LOG_ERROR, "pairparsevalue: Unknown value %s for attribute %s\n", value, vp->name);
				return NULL;
			} else {
				vp->lvalue = dval->value;
				vp->length = 4;
			}
			break;

		case PW_TYPE_DATE:
			if (gettime(value, (time_t *)&vp->lvalue) < 0) {
				ast_log(LOG_ERROR, "pairparsevalue: failed to parse time string \"%s\"\n", value);
				return NULL;
			}
			vp->length = 4;
			break;
		case PW_TYPE_ABINARY:
#ifdef ASCEND_BINARY
			/*
			 *	Special case to convert filter to binary
			 */
		  	if ( filterBinary( vp, value ) < 0 ) {
				ast_log(LOG_ERROR, "pairparsevalue: failed to parse Ascend binary attribute: %s\n", librad_errstr);
			  	return NULL;
			}
			break;
			/*
			 *	If Ascend binary is NOT defined,
			 *	then fall through to raw octets, so that
			 *	the user can at least make them by hand...
			 */
#endif
			/* raw octets: 0x01020304... */
		case PW_TYPE_OCTETS:
			if (strncasecmp(value, "0x", 2) == 0) {
				u_char *us;
				cp = value + 2;
				us = vp->strvalue;
				vp->length = 0;

				while (*cp && vp->length < MAX_STRING_LEN) {
					unsigned int tmp;
					
					if (sscanf(cp, "%02x", &tmp) != 1) {
						ast_log(LOG_ERROR, "pairparsevalue: Non-hex characters at %c%c", cp[0], cp[1]);
						return NULL;
					}

					cp += 2;
					*(us++) = tmp;
					vp->length++;
				}
				*us = '\0';
			}
			break;

			/*
			 *  Anything else.
			 */
		default:
			ast_log(LOG_ERROR, "pairparsevalue: unknown attribute type %d", vp->type);
			return NULL;
	}

	return vp;
}

VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
{
	DICT_ATTR	*da;
	VALUE_PAIR	*vp;
	char            *tc, *ts;
	signed char     tag;
	int             found_tag;

	/*
	 *    Check for tags in 'Attribute:Tag' format.
	 */
	found_tag = 0;
	tag = 0;

	ts = strrchr( attribute, ':' );
	if (ts && ts[1]) {
	         /* Colon found with something behind it */
	         if (ts[1] == '*' && ts[2] == 0) {
		         /* Wildcard tag for check items */
		         tag = TAG_ANY;
			 *ts = 0;
		 } else if ((ts[1] >= 0) && (ts[1] <= '9')) {
		         /* It's not a wild card tag */
		         tag = strtol(ts + 1, &tc, 0);
			 if (tc && !*tc && TAG_VALID_ZERO(tag)) {
			 	*ts = 0;
			} else {
				tag = 0;
			}
		 } else {
			 ast_log(LOG_ERROR, "pairmake: Invalid tag for attribute %s", attribute);
			 return NULL;
		 }
		 found_tag = 1;
	}

	if ((da = dict_attrbyname(attribute)) == NULL) {
		ast_log(LOG_ERROR, "pairmake: Unknown attribute %s", attribute);
		return NULL;
	}

	if ((vp = (VALUE_PAIR *)malloc(sizeof(VALUE_PAIR))) == NULL) {
		ast_log(LOG_ERROR, "pairmake: out of memory");
		return NULL;
	}

	memset(vp, 0, sizeof(VALUE_PAIR));
	vp->attribute = da->attr;
	vp->type = da->type;
	vp->operator = (operator == 0) ? T_OP_EQ : operator;
	strcpy(vp->name, da->name);
	vp->flags = da->flags;
	vp->next = NULL;

	/*      Check for a tag in the 'Merit' format of:
	 *      :Tag:Value.  Print an error if we already found
	 *      a tag in the Attribute.
	 */

	if (value && (*value == ':' && da->flags.has_tag)) {
	        /* If we already found a tag, this is invalid */
	        if(found_tag) {
		        free(vp);
			ast_log(LOG_ERROR, "pairmake: Duplicate tag %s for attribute %s", value, vp->name);
			return NULL;

		}
	        /* Colon found and attribute allows a tag */
	        if (value[1] == '*' && value[2] == ':') {
		       /* Wildcard tag for check items */
		       tag = TAG_ANY;
		       value += 3;
		} else {
	               /* Real tag */
		       tag = strtol(value + 1, &tc, 0);
		       if (tc && *tc==':' && TAG_VALID_ZERO(tag))
			    value = tc + 1;
		       else tag = 0;
		}
		found_tag = 1;
	}	
	
	if (found_tag) {
	  vp->flags.tag = tag;
	}

	switch (vp->operator) {
	default:
		break;

		/*
		 *      For =* and !* operators, the value is irrelevant
		 *      so we return now.
		 */
	case T_OP_CMP_TRUE:
	case T_OP_CMP_FALSE:
		vp->strvalue[0] = '\0';
		vp->length = 0;
	        return vp;
		break;

#ifdef HAVE_REGEX_H
		/*
		 *	Regular expression comparison of integer attributes
		 *	does a STRING comparison of the names of their
		 *	integer attributes.
		 */
	case T_OP_REG_EQ:	/* =~ */
	case T_OP_REG_NE:	/* !~ */
		if (vp->type == PW_TYPE_INTEGER) {
			return vp;
		}
#endif
	}

	if (value && (pairparsevalue(vp, value) == NULL)) {
		free(vp);
		return NULL;
	}

	return vp;
}

VALUE_PAIR *pairread(char **ptr, LRAD_TOKEN *eol)
{
	char		buf[64];
	char		attr[64];
	char		value[256];
	char		*p;
	LRAD_TOKEN	token, t, xlat;
	VALUE_PAIR	*vp;

	*eol = 0;

	/* Get attribute. */
	token = gettoken(ptr, attr, sizeof(attr));

	/*  If it's a comment, then exit, as we haven't read a pair */
	if (token == T_HASH) {
		*eol = token;
		ast_log(LOG_ERROR, "pairread: Read a comment instead of a token");
		return NULL;
	}

	/*  It's not a comment, so it MUST be an attribute */
	if ((token == T_EOL) ||
	    (attr[0] == 0)) {
		ast_log(LOG_ERROR, "pairread: No token read where we expected an attribute name");
		return NULL;
	}

	/* Now we should have an '=' here. */
	token = gettoken(ptr, buf, sizeof(buf));
	if (token < T_EQSTART || token > T_EQEND) {
		ast_log(LOG_ERROR, "pairread: expecting '='");
		return NULL;
	}

	/* Read value.  Note that empty string values are allowed */
	xlat = gettoken(ptr, value, sizeof(value));
	if (xlat == T_EOL) {
		ast_log(LOG_ERROR, "pairread: failed to get value");
		return NULL;
	}

	/*
	 *	Peek at the next token. Must be T_EOL, T_COMMA, or T_HASH
	 */
	p = *ptr;
	t = gettoken(&p, buf, sizeof(buf));
	if (t != T_EOL && t != T_COMMA && t != T_HASH) {
		ast_log(LOG_ERROR, "pairread: Expected end of line or comma");
		return NULL;
	}

	*eol = t;
	if (t == T_COMMA) {
		*ptr = p;
	}

	switch (xlat) {
		/*
		 *	Make the full pair now.
		 */
	default:
		vp = pairmake(attr, value, token);
		break;

		/*
		 *	Mark the pair to be allocated later.
		 */
	case T_BACK_QUOTED_STRING:
		vp = pairmake(attr, NULL, token);
		if (!vp) return vp;
		
		vp->flags.do_xlat = 1;
		strncpy(vp->strvalue, value, sizeof(vp->strvalue));
		vp->length = 0;
		break;
	}
	
	return vp;
}

LRAD_TOKEN userparse(char *buffer, VALUE_PAIR **first_pair)
{
	VALUE_PAIR	*vp;
	char		*p;
	LRAD_TOKEN	last_token = T_INVALID;
	LRAD_TOKEN	previous_token;

	/*
	 *	We allow an empty line.
	 */
	if (buffer[0] == 0)
		return T_EOL;

	p = buffer;
	do {
		previous_token = last_token;
		if ((vp = pairread(&p, &last_token)) == NULL) {
			return T_INVALID;
		}
		pairadd(first_pair, vp);
	} while (*p && (last_token == T_COMMA));

	/*
	 *	Don't tell the caller that there was a comment.
	 */
	if (last_token == T_HASH) {
		return previous_token;
	}

	/*
	 *	And return the last token which we read.
	 */
	return last_token;
}

static const char *dtypes[] = {
	"string",
	"integer",
	"ipaddr",
	"date",
	"abinary",
	"octets",
	NULL,
};

void pairfree(VALUE_PAIR **pair_ptr)
{
	VALUE_PAIR	*next, *pair;

	if (!pair_ptr) return;
	pair = *pair_ptr;
	while (pair != NULL) {
		next = pair->next;
		free(pair);
		pair = next;
	}
	*pair_ptr = NULL;
}

static void make_secret(unsigned char *digest, uint8_t *vector, const char *secret, char *value)
{
        u_char	buffer[256 + AUTH_VECTOR_LEN];
        int	secretLen = strlen(secret);
        int	i;

        memcpy(buffer, vector, AUTH_VECTOR_LEN );
        memcpy(buffer + AUTH_VECTOR_LEN, secret, secretLen );

        librad_md5_calc(digest, buffer, AUTH_VECTOR_LEN + secretLen );
        memset(buffer, 0, sizeof(buffer));
        for ( i = 0; i < AUTH_VECTOR_LEN; i++ ) {
		digest[i] ^= value[i];
        }
}

static void random_vector(uint8_t *vector)
{
	int		i;
	static int	did_srand = 0;
	static int	counter = 0;
	static int	urandom_fd = -1;

	/*
	 *	Use /dev/urandom if available.
	 */
	if (urandom_fd > -2) {
		if (urandom_fd < 0)
			urandom_fd = open("/dev/urandom", O_RDONLY);
		if (urandom_fd < 0) {
			ast_log(LOG_WARNING,"random_vector: Cannot open /dev/urandom, using rand()\n");
			urandom_fd = -2;
		} else {
			fcntl(urandom_fd, F_SETFD, 1);
			if (read(urandom_fd, (char *) vector, AUTH_VECTOR_LEN) == AUTH_VECTOR_LEN) return;
			ast_log(LOG_WARNING, "random_vector: Read short packet from /dev/urandom, using rand()\n");
			urandom_fd = -2;
		}
	}

	if (!did_srand) {
		srand(time(NULL) + getpid());

		/*
		 *	Now that we have a bad random seed, let's
		 *	make it a little better by MD5'ing it.
		 */
		for (i = 0; i < (int)sizeof(random_vector_pool); i++) {
			random_vector_pool[i] += rand() & 0xff;
		}

		librad_md5_calc((u_char *) random_vector_pool, (u_char *) random_vector_pool, sizeof(random_vector_pool));
		did_srand = 1;
	}

	/*
	 *	Modify our random pool, based on the counter,
	 *	and put the resulting information through MD5,
	 *	so it's all mashed together.
	 */
	counter++;
	random_vector_pool[AUTH_VECTOR_LEN] += (counter & 0xff);
	librad_md5_calc((u_char *) random_vector_pool, (u_char *) random_vector_pool, sizeof(random_vector_pool));

	/*
	 *	And do another MD5 hash of the result, to give
	 *	the user a random vector.  This ensures that the
	 *	user has a random vector, without giving them
	 *	an exact image of what's in the random pool.
	 */
	librad_md5_calc((u_char *) vector, (u_char *) random_vector_pool, sizeof(random_vector_pool));
}

RADIUS_PACKET *rad_alloc(int newvector)
{
	RADIUS_PACKET	*rp;

	rp = malloc(sizeof(RADIUS_PACKET)); 
	if (!rp) {
		ast_log(LOG_ERROR, "rad_alloc: out of memory");
		return NULL;
	}
	memset(rp, 0, sizeof(RADIUS_PACKET));
	if (newvector) random_vector(rp->vector);
	return rp;
}

void rad_free(RADIUS_PACKET **radius_packet_ptr)
{
	RADIUS_PACKET *radius_packet;

	if (!radius_packet_ptr) return;
	radius_packet = *radius_packet_ptr;

	if (radius_packet->data) free(radius_packet->data);
	if (radius_packet->vps) pairfree(&radius_packet->vps);
	free(radius_packet);
	*radius_packet_ptr = NULL;
}

int rad_pwencode(char *passwd, int *pwlen, const char *secret, const char *vector)
{
	uint8_t	buffer[AUTH_VECTOR_LEN + MAX_STRING_LEN + 1];
	char	digest[AUTH_VECTOR_LEN];
	int	i, n, secretlen;
	int	len;

	/*
	 *	Padd password to multiple of AUTH_PASS_LEN bytes.
	 */
	len = strlen(passwd);
	if (len > 128) len = 128;
	*pwlen = len;
	if (len % AUTH_PASS_LEN != 0) {
		n = AUTH_PASS_LEN - (len % AUTH_PASS_LEN);
		for (i = len; n > 0; n--, i++)
			passwd[i] = 0;
		len = *pwlen = i;
	}

	/*
	 *	Use the secret to setup the decryption digest
	 */
	secretlen = strlen(secret);
	memcpy(buffer, secret, secretlen);
	memcpy(buffer + secretlen, vector, AUTH_VECTOR_LEN);
	librad_md5_calc((u_char *)digest, buffer, secretlen + AUTH_VECTOR_LEN);

	/*
	 *	Now we can encode the password *in place*
	 */
	for (i = 0; i < AUTH_PASS_LEN; i++)
		passwd[i] ^= digest[i];

	if (len <= AUTH_PASS_LEN) return 0;

	/*
	 *	Length > AUTH_PASS_LEN, so we need to use the extended
	 *	algorithm.
	 */
	for (n = 0; n < 128 && n <= (len - AUTH_PASS_LEN); n += AUTH_PASS_LEN) { 
		memcpy(buffer + secretlen, passwd + n, AUTH_PASS_LEN);
		librad_md5_calc((u_char *)digest, buffer, secretlen + AUTH_PASS_LEN);
		for (i = 0; i < AUTH_PASS_LEN; i++)
			passwd[i + n + AUTH_PASS_LEN] ^= digest[i];
	}

	return 0;
}

int rad_pwdecode(char *passwd, int pwlen, const char *secret, const char *vector)
{
	uint8_t	buffer[AUTH_VECTOR_LEN + MAX_STRING_LEN + 1];
	char	digest[AUTH_VECTOR_LEN];
	char	r[AUTH_VECTOR_LEN];
	char	*s;
	int	i, n, secretlen;
	int	rlen;

	/*
	 *	Use the secret to setup the decryption digest
	 */
	secretlen = strlen(secret);
	memcpy(buffer, secret, secretlen);
	memcpy(buffer + secretlen, vector, AUTH_VECTOR_LEN);
	librad_md5_calc((u_char *)digest, buffer, secretlen + AUTH_VECTOR_LEN);

	/*
	 *	Now we can decode the password *in place*
	 */
	memcpy(r, passwd, AUTH_PASS_LEN);
	for (i = 0; i < AUTH_PASS_LEN && i < pwlen; i++)
		passwd[i] ^= digest[i];

	if (pwlen <= AUTH_PASS_LEN) {
		passwd[pwlen+1] = 0;
		return pwlen;
	}

	/*
	 *	Length > AUTH_PASS_LEN, so we need to use the extended
	 *	algorithm.
	 */
	rlen = ((pwlen - 1) / AUTH_PASS_LEN) * AUTH_PASS_LEN;

	for (n = rlen; n > 0; n -= AUTH_PASS_LEN ) { 
		s = (n == AUTH_PASS_LEN) ? r : (passwd + n - AUTH_PASS_LEN);
		memcpy(buffer + secretlen, s, AUTH_PASS_LEN);
		librad_md5_calc((u_char *)digest, buffer, secretlen + AUTH_PASS_LEN);
		for (i = 0; i < AUTH_PASS_LEN && (i + n) < pwlen; i++)
			passwd[i + n] ^= digest[i];
	}
	passwd[pwlen] = 0;

	return pwlen;
}

int rad_tunnel_pwencode(char *passwd, int *pwlen, const char *secret, const char *vector)
{
	uint8_t	buffer[AUTH_VECTOR_LEN + MAX_STRING_LEN + 3];
	unsigned char	digest[AUTH_VECTOR_LEN];
	char*   salt;
	int	i, n, secretlen;
	unsigned len;

	len = *pwlen;

	if (len > 127) len = 127;
	/*
	 * Shift the password 3 positions right to place a salt and original
	 * length, tag will be added automatically on packet send
	 */
	for (n=len ; n>=0 ; n--) passwd[n+3] = passwd[n];
	salt = passwd;
	passwd += 2;
	/*
	 * save original password length as first password character;
	 */
	*passwd = len;
	len += 1;


	/* generate salt */
	saltoffset++;
	salt[0] = (vector[0] ^ vector[1] ^ 0x3A ^ (char) ( (saltoffset>>8) & 0x000000ff ) ) | 0x80;
	salt[1] = (vector[2] ^ vector[3] ^ vector[4] ^ (char) (saltoffset & 0x000000ff) );
	
	/*
	 *	Padd password to multiple of AUTH_PASS_LEN bytes.
	 */
	n = len % AUTH_PASS_LEN;
	if (n) {
		n = AUTH_PASS_LEN - n;
		for (; n > 0; n--, len++)
			passwd[len] = 0;
	}
	/* set new password length */
	*pwlen = len + 2;
	
	/*
	 *	Use the secret to setup the decryption digest
	 */
	secretlen = strlen(secret);
	memcpy(buffer, secret, secretlen);
	
	for (n = 0; n < len; n+=AUTH_PASS_LEN) {
		if (!n) {
			memcpy(buffer + secretlen, vector, AUTH_VECTOR_LEN);
			memcpy(buffer + secretlen + AUTH_VECTOR_LEN, salt, 2);
			librad_md5_calc(digest, buffer, secretlen + AUTH_VECTOR_LEN + 2);
		}
		else {
			memcpy(buffer + secretlen, passwd + n - AUTH_PASS_LEN, AUTH_PASS_LEN);
			librad_md5_calc(digest, buffer, secretlen + AUTH_PASS_LEN);
		}
		
		for (i = 0; i < AUTH_PASS_LEN; i++)
			passwd[i + n] ^= digest[i];
	}
	passwd[n] = 0;
	return 0;
}

int rad_tunnel_pwdecode(char *passwd, int * pwlen, const char *secret, const char *vector)
{
	uint8_t	buffer[AUTH_VECTOR_LEN + MAX_STRING_LEN + 3];
	unsigned char	digest[AUTH_VECTOR_LEN];
	char    salt[2];
	int	i, n, ntimes, secretlen;
	unsigned len;
	
	len = *pwlen;

	if(len < 3) {
	  return len;
	}
	salt[0] = passwd[0];
	salt[1] = passwd[1];

	passwd += 2;
	len -= 2;
	
	/*
	 *	Use the secret to setup the decryption digest
	 */
	secretlen = strlen(secret);
	memcpy(buffer, secret, secretlen);

	ntimes = (len-1)/AUTH_PASS_LEN;
	do {
		if(!ntimes){
			memcpy(buffer + secretlen, vector, AUTH_VECTOR_LEN);
			memcpy(buffer + secretlen + AUTH_VECTOR_LEN, salt, 2);
			librad_md5_calc(digest, buffer, secretlen + AUTH_VECTOR_LEN + 2);
		}
		else {
			memcpy(buffer + secretlen, passwd + AUTH_PASS_LEN * (ntimes - 1), AUTH_PASS_LEN);
			librad_md5_calc(digest, buffer, secretlen + AUTH_PASS_LEN);
		}
		for ( i = 0, n = ntimes * AUTH_PASS_LEN; i < AUTH_PASS_LEN && (i + n) < len; i++)
			passwd[i + n] ^= digest[i];
	} while(ntimes--);
	passwd[len] = '\0';

	if (*(unsigned char*)passwd >= len) {
		/* Pasword is broken, original password should be longer */
		*pwlen = 2;
		passwd[0]=passwd[1]=0;
		return 0;
	}
	len = *pwlen = *passwd; /* restore original length */
	for (n=0; n<len; n++) passwd[n-2]=passwd[n+1];
	passwd[len-2] = 0;
	return len;
}

int rad_decode(RADIUS_PACKET *packet, RADIUS_PACKET *original, const char *secret)
{
	DICT_ATTR		*attr;
	uint32_t		lvalue;
	uint32_t		vendorcode;
	VALUE_PAIR		**tail;
	VALUE_PAIR		*pair;
	uint8_t			*ptr;
	int			length;
	int			attribute;
	int			attrlen;
	int			vendorlen;
	radius_packet_t		*hdr;

	hdr = (radius_packet_t *)packet->data;

	/*
	 *	Before we allocate memory for the attributes, do more
	 *	sanity checking.
	 */
	ptr = hdr->data;
	length = packet->data_len - AUTH_HDR_LEN;
	while (length > 0) {
		uint8_t	msg_auth_vector[AUTH_VECTOR_LEN];
		uint8_t calc_auth_vector[AUTH_VECTOR_LEN];

		attrlen = ptr[1];
		switch (ptr[0]) {
			default:	/* don't do anything. */
				break;

				/*
				 *	Note that more than one Message-Authenticator
				 *	attribute is invalid.
				 */
			case PW_MESSAGE_AUTHENTICATOR:
				memcpy(msg_auth_vector, &ptr[2], sizeof(msg_auth_vector));
				memset(&ptr[2], 0, AUTH_VECTOR_LEN);

				switch (packet->code) {
					default:
						break;

					case PW_AUTHENTICATION_ACK:
					case PW_AUTHENTICATION_REJECT:
					case PW_ACCESS_CHALLENGE:
						if (original) {
							memcpy(packet->data + 4, original->vector, AUTH_VECTOR_LEN);
						}
						break;
				}

				lrad_hmac_md5(packet->data, packet->data_len, secret, strlen(secret), calc_auth_vector);
				if (memcmp(calc_auth_vector, msg_auth_vector, sizeof(calc_auth_vector)) != 0) {
					char buffer[32];
					ast_log(LOG_ERROR,"rad_decode: Received packet from %s with invalid Message-Authenticator!", ip_ntoa(buffer, packet->src_ipaddr));
					return 1;
				} /* else the message authenticator was good */

				/*
				 *	Reinitialize Authenticators.
				 */
				memcpy(&ptr[2], msg_auth_vector, AUTH_VECTOR_LEN);
				memcpy(packet->data + 4, packet->vector, AUTH_VECTOR_LEN);
				break;
		} /* switch over the attributes */

		ptr += attrlen;
		length -= attrlen;
	} /* loop over the packet, sanity checking the attributes */

	/*
	 *	Calculate and/or verify digest.
	 */
	switch(packet->code) {
		case PW_AUTHENTICATION_REQUEST:
		case PW_STATUS_SERVER:
			/*
			 *	The authentication vector is random
			 *	nonsense, invented by the client.
			 */
			break;

		case PW_ACCOUNTING_REQUEST:
			if (calc_acctdigest(packet, secret) > 1) {
				char buffer[32];
				ast_log(LOG_ERROR, "Received Accounting-Request packet from %s with invalid signature!", ip_ntoa(buffer, packet->src_ipaddr));
				return 1;
			}
			break;

			/* Verify the reply digest */
		case PW_AUTHENTICATION_ACK:
		case PW_AUTHENTICATION_REJECT:
		case PW_ACCOUNTING_RESPONSE:
			if (calc_replydigest(packet, original, secret) > 1) {
				char buffer[32];
				ast_log(LOG_ERROR, "Received %s packet from %s with invalid signature!", packet_codes[packet->code], ip_ntoa(buffer, packet->src_ipaddr));
				return 1;
			}
			break;
	}

	/*
	 *	Extract attribute-value pairs
	 */
	ptr = hdr->data;
	length = packet->data_len - AUTH_HDR_LEN;
	packet->vps = NULL;
	tail = &packet->vps;

	vendorcode = 0;
	vendorlen  = 0;

	while(length > 0) {
		if (vendorlen > 0) {
			attribute = *ptr++ | (vendorcode << 16);
			attrlen   = *ptr++;
		} else {
			attribute = *ptr++;
			attrlen   = *ptr++;
		}

		attrlen -= 2;
		length  -= 2;

		/*
		 *	This could be a Vendor-Specific attribute.
		 *
		 */
		if ((vendorlen <= 0) && (attribute == PW_VENDOR_SPECIFIC) && (attrlen > 6)) {
			memcpy(&lvalue, ptr, 4);
			vendorcode = ntohl(lvalue);
			if (vendorcode != 0) {
				if (vendorcode == VENDORPEC_USR) {
					ptr += 4;
					memcpy(&lvalue, ptr, 4);
					/*printf("received USR %04x\n", ntohl(lvalue));*/
					attribute = (ntohl(lvalue) & 0xFFFF) | (vendorcode << 16);
					ptr += 4;
					attrlen -= 8;
					length -= 8;

				} else {
					ptr += 4;
					vendorlen = attrlen - 4;
					attribute = *ptr++ | (vendorcode << 16);
					attrlen   = *ptr++;
					attrlen -= 2;
					length -= 6;
				}
			}
		}

		/*
		 *	FIXME: should we us paircreate() ?
		 */
		if ((pair = malloc(sizeof(VALUE_PAIR))) == NULL) {
			pairfree(&packet->vps);
			ast_log(LOG_ERROR, "rad_decode: out of memory");
			errno = ENOMEM;
			return -1;
		}
		
		memset(pair, 0, sizeof(VALUE_PAIR));
		if ((attr = dict_attrbyvalue(attribute)) == NULL) {
			snprintf(pair->name, sizeof(pair->name), "Attr-%d", attribute);
			pair->type = PW_TYPE_OCTETS;
		} else {
			strcpy(pair->name, attr->name);
			pair->type = attr->type;
			pair->flags = attr->flags;
		}
		pair->attribute = attribute;
		pair->length = attrlen;
		pair->operator = T_OP_EQ;
		pair->next = NULL;

		switch (pair->type) {
			case PW_TYPE_OCTETS:
			case PW_TYPE_ABINARY:
			case PW_TYPE_STRING:
				if (pair->flags.has_tag && pair->type == PW_TYPE_STRING) {
					int offset = 0;
					if(TAG_VALID(*ptr)) {
						pair->flags.tag = *ptr;
						pair->length--;
						offset = 1;
					} else if (pair->flags.encrypt == FLAG_ENCRYPT_TUNNEL_PASSWORD) {
						/*
						 * from RFC2868 - 3.5.  Tunnel-Password
						 * If the value of the Tag field is greater than
						 * 0x00 and less than or equal to 0x1F, it SHOULD
						 * be interpreted as indicating which tunnel
						 * (of several alternatives) this attribute pertains;
						 * otherwise, the Tag field SHOULD be ignored.
						 */
						pair->flags.tag = 0x00;
						pair->length--;
						offset = 1;
					} else {
						pair->flags.tag = 0x00;
					}
					memcpy(pair->strvalue, ptr + offset, pair->length);
				} else {
					/* attrlen always < MAX_STRING_LEN */
					memcpy(pair->strvalue, ptr, attrlen);
				        pair->flags.tag = 0;
				}

				/*
				 *  FIXME: HACK for non-updated dictionaries.
				 *  REMOVE in a future release.
				 */
				if (!strcmp(pair->name, "Ascend-Send-Secret") || !strcmp(pair->name, "Ascend-Receive-Secret")) {
					pair->flags.encrypt = FLAG_ENCRYPT_ASCEND_SECRET;
				}
				if (pair->attribute == PW_USER_PASSWORD) {
					pair->flags.encrypt = FLAG_ENCRYPT_USER_PASSWORD;
				}

				/*
				 *	Decrypt passwords here.
				 */
				switch (pair->flags.encrypt) {
					default:
						break;

					/*
					 *  User-Password
					 */
					case FLAG_ENCRYPT_USER_PASSWORD:
						rad_pwdecode((char *)pair->strvalue, pair->length, secret, (char *)packet->vector);
						pair->lvalue = 1; /* see main/auth.c */

					/*
					 *  FIXME: Allow NUL's in the password,
					 *  too??
					 */
					pair->length = strlen(pair->strvalue);
					break;

					/*
					 *  Tunnel-Password
					 */
					case FLAG_ENCRYPT_TUNNEL_PASSWORD:
			        		rad_tunnel_pwdecode((char *)pair->strvalue, &pair->length, secret, (char *)original->vector);
						break;

					/*
					 *  Ascend-Send-Secret
					 *  Ascend-Receive-Secret
					 */
					case FLAG_ENCRYPT_ASCEND_SECRET:
						{
							uint8_t my_digest[AUTH_VECTOR_LEN];
							make_secret(my_digest, original->vector, secret, ptr);
							memcpy(pair->strvalue, my_digest, AUTH_VECTOR_LEN );
							pair->strvalue[AUTH_VECTOR_LEN] = '\0';
							pair->length = strlen(pair->strvalue);
							break;
						}
				} /* switch over encryption flags */
				break;	/* from octets/string/abinary */
			
			case PW_TYPE_INTEGER:
			case PW_TYPE_DATE:
			case PW_TYPE_IPADDR:
				/*
				 *	Check for RFC compliance.  If the attribute isn't compliant, turn it into a string of raw octets.
				 *	Also set the lvalue to something which should never match anything.
				 */
				if (attrlen != 4) {
					pair->type = PW_TYPE_OCTETS;
					memcpy(pair->strvalue, ptr, attrlen);
					pair->lvalue = 0xbad1bad1;
					break;
				}
      				memcpy(&lvalue, ptr, 4);
				if (attr->type != PW_TYPE_IPADDR) {
					pair->lvalue = ntohl(lvalue);
				} else {
					 /*
					  *  It's an IP address, keep it in network byte order, and put the ASCII IP
					  *  address or host name into the string value.
					  */
					pair->lvalue = lvalue;
					ip_ntoa(pair->strvalue, pair->lvalue);
				}

				/*
				 *  Only PW_TYPE_INTEGER should have tags.
				 */
				if (pair->flags.has_tag && pair->type == PW_TYPE_INTEGER) {
				        pair->flags.tag = (pair->lvalue >> 24) & 0xff;
					pair->lvalue &= 0x00ffffff;
				}
				if (attr->type == PW_TYPE_INTEGER) {
					DICT_VALUE *dval;
					dval = dict_valbyattr(pair->attribute, pair->lvalue);
					if (dval) {
						strncpy(pair->strvalue, dval->name, sizeof(pair->strvalue));
					}
				}
				break;
			
			default:
				ast_log(LOG_WARNING, "rad_decode: %s (Unknown Type %d)\n", attr->name,attr->type);
				free(pair);
				pair = NULL;
				break;
		}
		
		if (pair) {
			*tail = pair;
			tail = &pair->next;
		}

		ptr += attrlen;
		length -= attrlen;
		if (vendorlen > 0) vendorlen -= (attrlen + 2);
	}

	/*
	 *	Merge information from the outside world into our
	 *	random vector pool.
	 */
	for (length = 0; length < AUTH_VECTOR_LEN; length++) {
		random_vector_pool[length] ^= packet->vector[length];
	}

	return 0;
}

RADIUS_PACKET *rad_recv(int fd)
{
	RADIUS_PACKET		*packet;
	struct sockaddr_in	saremote;
	int			totallen;
	socklen_t		salen;
	u_short			len;
	uint8_t			*attr;
	uint8_t			*vendorattr;
	int			count;
	radius_packet_t		*hdr;
	char			host_ipaddr[16];
	int			seen_eap;
	uint8_t			data[MAX_PACKET_LEN];
	int			num_attributes;
	uint32_t                vendorcode;
	int			vendorlen;
	
	/*
	 *	Allocate the new request data structure
	 */
	if ((packet = malloc(sizeof(RADIUS_PACKET))) == NULL) {
		ast_log(LOG_ERROR, "rad_recv: out of memory");
		errno = ENOMEM;
		return NULL;
	}
	memset(packet, 0, sizeof(RADIUS_PACKET));

	/*
	 *	Receive the packet.
	 */
	salen = sizeof(saremote);
	memset(&saremote, 0, sizeof(saremote));
	packet->data_len = recvfrom(fd, data, sizeof(data), 0, (struct sockaddr *)&saremote, &salen);

	/*
	 *	Check for socket errors.
	 */
	if (packet->data_len < 0) {
		ast_log(LOG_ERROR, "rad_recv: Error receiving packet: %s", strerror(errno));
		free(packet);
		return NULL;
	}

	/*
	 *	Fill IP header fields.  We need these for the error
	 *	messages which may come later.
	 */
	packet->sockfd = fd;
	packet->src_ipaddr = saremote.sin_addr.s_addr;
	packet->src_port = ntohs(saremote.sin_port);

	/*
	 *	Explicitely set the VP list to empty.
	 */
	packet->vps = NULL;

	/*
	 *	Check for packets smaller than the packet header.
	 *
	 *	RFC 2865, Section 3., subsection 'length' says:
	 *
	 *	"The minimum length is 20 ..."
	 */
	if (packet->data_len < AUTH_HDR_LEN) {
		ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: too short (received %d < minimum %d)", ip_ntoa(host_ipaddr, packet->src_ipaddr), packet->data_len, AUTH_HDR_LEN);
		free(packet);
		return NULL;
	}

	/*
	 *	RFC 2865, Section 3., subsection 'length' says:
	 *
	 *	" ... and maximum length is 4096."
	 */
	if (packet->data_len > MAX_PACKET_LEN) {
		ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: too long (received %d > maximum %d)", ip_ntoa(host_ipaddr, packet->src_ipaddr), packet->data_len, MAX_PACKET_LEN);
		free(packet);
		return NULL;
	}

	/*
	 *	Check for packets with mismatched size.
	 *	i.e. We've received 128 bytes, and the packet header
	 *	says it's 256 bytes long.
	 */
	hdr = (radius_packet_t *)data;
	memcpy(&len, hdr->length, sizeof(u_short));
	totallen = ntohs(len);

	/*
	 *	Code of 0 is not understood.
	 *	Code of 16 or greate is not understood.
	 */
	if ((hdr->code == 0) || (hdr->code >= 16)) {
		ast_log(LOG_ERROR, "rad_recv: Bad RADIUS packet from host %s: unknown packet code %d", ip_ntoa(host_ipaddr, packet->src_ipaddr), hdr->code);
		free(packet);
		return NULL;
		
	}

	/*
	 *	Repeat the length checks.  This time, instead of
	 *	looking at the data we received, look at the value
	 *	of the 'length' field inside of the packet.
	 *
	 *	Check for packets smaller than the packet header.
	 *
	 *	RFC 2865, Section 3., subsection 'length' says:
	 *
	 *	"The minimum length is 20 ..."
	 */
	if (totallen < AUTH_HDR_LEN) {
		ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: too short (length %d < minimum %d)", ip_ntoa(host_ipaddr, packet->src_ipaddr), totallen, AUTH_HDR_LEN);
		free(packet);
		return NULL;
	}

	/*
	 *	And again, for the value of the 'length' field.
	 *
	 *	RFC 2865, Section 3., subsection 'length' says:
	 *
	 *	" ... and maximum length is 4096."
	 */
	if (totallen > MAX_PACKET_LEN) {
		ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: too long (length %d > maximum %d)", ip_ntoa(host_ipaddr, packet->src_ipaddr), totallen, MAX_PACKET_LEN);
		free(packet);
		return NULL;
	}

	/*
	 *	RFC 2865, Section 3., subsection 'length' says:
	 *
	 *	"If the packet is shorter than the Length field
	 *	indicates, it MUST be silently discarded."
	 *
	 *	i.e. No response to the NAS.
	 */
	if (packet->data_len < totallen) {
		ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: received %d octets, packet length says %d", ip_ntoa(host_ipaddr, packet->src_ipaddr), packet->data_len, totallen);
		free(packet);
		return NULL;
	}

	/*
	 *	RFC 2865, Section 3., subsection 'length' says:
	 *
	 *	"Octets outside the range of the Length field MUST be
	 *	treated as padding and ignored on reception."
	 */
	if (packet->data_len > totallen) {
		/*
		 *	We're shortening the packet below, but just
		 *	to be paranoid, zero out the extra data.
		 */
		memset(data + totallen, 0, packet->data_len - totallen);
		packet->data_len = totallen;
	}

	/*
	 *	Walk through the packet's attributes, ensuring that
	 *	they add up EXACTLY to the size of the packet.
	 *
	 *	If they don't, then the attributes either under-fill
	 *	or over-fill the packet.  Any parsing of the packet
	 *	is impossible, and will result in unknown side effects.
	 *
	 *	This would ONLY happen with buggy RADIUS implementations,
	 *	or with an intentional attack.  Either way, we do NOT want
	 *	to be vulnerable to this problem.
	 */
	attr = hdr->data;
	count = totallen - AUTH_HDR_LEN;
	seen_eap = 0;
	num_attributes = 0;

	while (count > 0) {
		/*
		 *	Attribute number zero is NOT defined.
		 */
		if (attr[0] == 0) {
			ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: Invalid attribute 0", ip_ntoa(host_ipaddr, packet->src_ipaddr));
			free(packet);
			return NULL;
		}
		
		/*
		 *	Attributes are at LEAST as long as the ID & length
		 *	fields.  Anything shorter is an invalid attribute.
		 */
       		if (attr[1] < 2) {
			ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: attribute %d too short", ip_ntoa(host_ipaddr, packet->src_ipaddr), attr[0]);
			free(packet);
			return NULL;
		}

		/*
		 *	Sanity check the attributes for length.
		 */
		switch (attr[0]) {
			default:	/* don't do anything by default */
				break;

			case PW_EAP_MESSAGE:
				seen_eap |= PW_EAP_MESSAGE;
				break;

			case PW_MESSAGE_AUTHENTICATOR:
				if (attr[1] != 2 + AUTH_VECTOR_LEN) {
					ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: Message-Authenticator has invalid length %d", ip_ntoa(host_ipaddr, packet->src_ipaddr), attr[1] - 2);
					free(packet);
					return NULL;
				}
				seen_eap |= PW_MESSAGE_AUTHENTICATOR;
				break;
			
			case PW_VENDOR_SPECIFIC:
				if (attr[1] < 6) {
					ast_log(LOG_ERROR,"rad_recv: Malformed RADIUS packet from host %s: Vendor-Specific has invalid length %d",
						   ip_ntoa(host_ipaddr, packet->src_ipaddr),
						   attr[1] - 2);
					free(packet);
					return NULL;
				}
				memcpy(&vendorcode, attr + 2, 4);
				vendorcode = ntohl(vendorcode);
				if (vendorcode == VENDORPEC_USR) {
					if (attr[1] < 8){
						ast_log(LOG_ERROR,"rad_recv: Malformed RADIUS packet from host %s: USR attribute has invalid length %d",
						   ip_ntoa(host_ipaddr, packet->src_ipaddr),
						   attr[1] - 2);
						free(packet);
						return NULL;
					}
					break;
				}
				vendorlen = attr[1] - 6;
				vendorattr = attr + 6;
				while (vendorlen >= 2) {
					if (vendorattr[1] < 2){
						ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: Vendor specific attribute has invalid length %d",
						   ip_ntoa(host_ipaddr, packet->src_ipaddr),
						   vendorattr[1] - 2);
						free(packet);
						return NULL;
					}
					vendorlen -= vendorattr[1];
					vendorattr += vendorattr[1];
				}
				if (vendorlen != 0){
					ast_log(LOG_ERROR, "WARNING: Malformed RADIUS packet from host %s: Vendor specific attributes do not exactly fill Vendor-Specific",
						   ip_ntoa(host_ipaddr, packet->src_ipaddr));
					free(packet);
					return NULL;
				}
				break;
		}

		/*
		 *	FIXME: Look up the base 255 attributes in the
		 *	dictionary, and switch over their type.  For
		 *	integer/date/ip, the attribute length SHOULD
		 *	be 6.
		 */
		count -= attr[1];	/* grab the attribute length */
		attr += attr[1];
		num_attributes++;	/* seen one more attribute */
	}

	/*
	 *	If the attributes add up to a packet, it's allowed.
	 *
	 *	If not, we complain, and throw the packet away.
	 */
	if (count != 0) {
		ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: packet attributes do NOT exactly fill the packet", ip_ntoa(host_ipaddr, packet->src_ipaddr));
		free(packet);
		return NULL;
	}

	/*
	 *	If we've seen an EAP-Message attribute without a
	 *	Message-Authenticator attribute, it's invalid.
	 */
	if (((seen_eap & PW_EAP_MESSAGE) == PW_EAP_MESSAGE) && ((seen_eap & PW_MESSAGE_AUTHENTICATOR) != PW_MESSAGE_AUTHENTICATOR)) {
		ast_log(LOG_ERROR, "rad_recv: Malformed RADIUS packet from host %s: Contains EAP-Message, but no Message-Authenticator",
			   ip_ntoa(host_ipaddr, packet->src_ipaddr));
		free(packet);
		return NULL;
	}

	/*
	 *	If we're configured to look for a maximum number of
	 *	attributes, and we've seen more than that maximum,
	 *	then throw the packet away, as a possible DoS.
	 */
	if ((librad_max_attributes > 0) && (num_attributes > librad_max_attributes)) {
		ast_log(LOG_ERROR, "rad_recv: Possible DoS attack from host %s: Too many attributes in request (received %d, max %d are allowed).",
			   ip_ntoa(host_ipaddr, packet->src_ipaddr),
			   num_attributes, librad_max_attributes);
		free(packet);
		return NULL;
	}

	/*
	 *	Fill RADIUS header fields
	 */
	packet->code = hdr->code;
	packet->id = hdr->id;
	memcpy(packet->vector, hdr->vector, AUTH_VECTOR_LEN);

	/*
	 *  Now that we've sanity checked the packet, we can allocate
	 *  memory for it, and copy the data from the local area to
	 *  the packet buffer.
	 */
	if ((packet->data = malloc(packet->data_len)) == NULL) {
		free(packet);
		ast_log(LOG_ERROR, "rad_recv: out of memory");
		return NULL;
	}
	memcpy(packet->data, data, packet->data_len);

	return packet;
}

int rad_send(RADIUS_PACKET *packet, const RADIUS_PACKET *original, const char *secret)
{
	VALUE_PAIR		*reply;
	struct	sockaddr_in	saremote;
	struct	sockaddr_in	*sa;
	const char		*what;
	uint8_t			ip_buffer[16];

	if ((packet->code > 0) && (packet->code < 14)) {
		what = packet_codes[packet->code];
	} else {
		what = "Reply";
	}

	/*
	 *  First time through, allocate room for the packet
	 */
	if (!packet->data) {
		radius_packet_t	*hdr;
		uint32_t		lvalue;
		uint8_t		*ptr, *length_ptr, *vsa_length_ptr;
		uint8_t		digest[16];
		int			secretlen;
		int			vendorcode, vendorpec;
		u_short		total_length;
		int			len, allowed;
		int			msg_auth_offset = 0;
		uint8_t		data[MAX_PACKET_LEN];
		
		/*
		 *	Use memory on the stack, until we know how
		 *	large the packet will be.
		 */
		hdr = (radius_packet_t *) data;

		/*
		 *	Build standard header
		 */
		hdr->code = packet->code;
		hdr->id = packet->id;
		if (packet->code == PW_ACCOUNTING_REQUEST) {
			memset(hdr->vector, 0, AUTH_VECTOR_LEN);
		} else {
			memcpy(hdr->vector, packet->vector, AUTH_VECTOR_LEN);
		}

		ast_log(LOG_DEBUG,"Sending %s of id %d to %s:%d\n", what, packet->id, ip_ntoa((char *)ip_buffer, packet->dst_ipaddr), packet->dst_port);
		
		total_length = AUTH_HDR_LEN;
		
		/*
		 *	Load up the configuration values for the user
		 */
		ptr = hdr->data;
		vendorcode = 0;
		vendorpec = 0;
		vsa_length_ptr = NULL;

		for (reply = packet->vps; reply; reply = reply->next) {
			/*
			 *	Ignore non-wire attributes
			 */
			if ((VENDOR(reply->attribute) == 0) && ((reply->attribute & 0xFFFF) > 0xff)) {
				continue;
			}

			/*
			 *	Do stuff for Message-Authenticator
			 */
			if (reply->attribute == PW_MESSAGE_AUTHENTICATOR) {
				/*
				 *  Set it to zero!
				 */
				reply->length = AUTH_VECTOR_LEN;
				memset(reply->strvalue, 0, AUTH_VECTOR_LEN);
				msg_auth_offset = total_length;
			}

			/*
			 *	We have a different vendor.  Re-set
			 *	the vendor codes.
			 */
			if (vendorcode != VENDOR(reply->attribute)) {
				vendorcode = 0;
				vendorpec = 0;
				vsa_length_ptr = NULL;
			}

			/*
			 *	If the Vendor-Specific attribute is getting
			 *	full, then create a new VSA attribute
			 *
			 *	FIXME: Multiple VSA's per Vendor-Specific
			 *	SHOULD be configurable.  When that's done,
			 *	the (1), below, can be changed to point to
			 *	a configuration variable which is set TRUE
			 *	if the NAS cannot understand multiple VSA's
			 *	per Vendor-Specific
			 */
			if ((1) || (vsa_length_ptr && (reply->length + *vsa_length_ptr) >= MAX_STRING_LEN)) {
				vendorcode = 0;
				vendorpec = 0;
				vsa_length_ptr = NULL;
			}

			/*
			 *	Maybe we have the start of a set of
			 *	(possibly many) VSA attributes from
			 *	one vendor.  Set a global VSA wrapper
			 */
			if ((vendorcode == 0) && ((vendorcode = VENDOR(reply->attribute)) != 0)) {
				vendorpec  = dict_vendorpec(vendorcode);
				
				/*
				 *	This is a potentially bad error...
				 *	we can't find the vendor ID!
				 */
				if (vendorpec == 0) {
					/* FIXME: log an error */
					continue;
				}

				/*
				 *	Build a VSA header.
				 */
				*ptr++ = PW_VENDOR_SPECIFIC;
				vsa_length_ptr = ptr;
				*ptr++ = 6;
				lvalue = htonl(vendorpec);
				memcpy(ptr, &lvalue, 4);
				ptr += 4;
				total_length += 6;
			}

			if (vendorpec == VENDORPEC_USR) {
				lvalue = htonl(reply->attribute & 0xFFFF);
				memcpy(ptr, &lvalue, 4);

				length_ptr = vsa_length_ptr;

				total_length += 4;
				*length_ptr  += 4;
				ptr          += 4;

				/*
				 *	Each USR attribute gets it's own
				 *	VSA wrapper, so we re-set the
				 *	vendor specific information.
				 */
				vendorcode = 0;
				vendorpec = 0;
				vsa_length_ptr = NULL;

			} else {
				/*
				 *	All other attributes are as
				 *	per the RFC spec.
				 */

				*ptr++ = (reply->attribute & 0xFF);
				length_ptr = ptr;
				if (vsa_length_ptr) *vsa_length_ptr += 2;
				*ptr++ = 2;
				total_length += 2;
			}
			
			switch(reply->type) {
				
				/*
				 *	Ascend binary attributes are
				 *	stored internally in binary form.
				 */
				case PW_TYPE_ABINARY:
				case PW_TYPE_STRING:
				case PW_TYPE_OCTETS:
					/*
					 *  FIXME: HACK for non-updated dictionaries.
					 *  REMOVE in a future release.
					 */
					if ((!strcmp(reply->name, "Ascend-Send-Secret")) || (!strcmp(reply->name, "Ascend-Receive-Secret"))) {
						reply->flags.encrypt = FLAG_ENCRYPT_ASCEND_SECRET;
					}
					if (reply->attribute == PW_USER_PASSWORD) {
						reply->flags.encrypt = FLAG_ENCRYPT_USER_PASSWORD;
					}

					/*
					 *  Encrypt the various password styles
					 */
					switch (reply->flags.encrypt) {
						default:
							break;

						case FLAG_ENCRYPT_USER_PASSWORD:
							rad_pwencode((char *)reply->strvalue, &(reply->length), (char *)secret, (char *)packet->vector);
							break;

						case FLAG_ENCRYPT_TUNNEL_PASSWORD:
							rad_tunnel_pwencode(reply->strvalue, &(reply->length), secret, packet->vector);
							break;

						case FLAG_ENCRYPT_ASCEND_SECRET:
							make_secret(digest, packet->vector, secret, reply->strvalue);
							memcpy(reply->strvalue, digest, AUTH_VECTOR_LEN );
							reply->length = AUTH_VECTOR_LEN;
					} /* switch over encryption flags */

					len = reply->length;

					/*
					 *    Set the TAG at the beginning
					 *    of the string if tagged.  If
					 *    tag value is not valid for
					 *    tagged attribute, make it 0x00
					 *    per RFC 2868.  -cparker
					 */
					if (reply->flags.has_tag) {
						if (TAG_VALID(reply->flags.tag)) {
							len++;
							*ptr++ = reply->flags.tag;
						} else if (reply->flags.encrypt == FLAG_ENCRYPT_TUNNEL_PASSWORD) {
							/*
							 *  Tunnel passwords REQUIRE a tag, even if we don't
							 *  have a valid tag.
							 */
							len++;
							*ptr++ = 0x00;
						} /* else don't write a tag */
					} /* else the attribute doesn't have a tag */
				 
					/*
					 *	Ensure we don't go too far. The 'length' of the attribute may be
					 *	0..255, minus whatever octets are used in the attribute header.
					 */
					allowed = 255;
					if (vsa_length_ptr) {
						allowed -= *vsa_length_ptr;
					} else {
						allowed -= *length_ptr;
					}
				
					if (len > allowed) {
						len = allowed;
					}
				
					*length_ptr += len;
					if (vsa_length_ptr) *vsa_length_ptr += len;
					/*
					 *  If we have tagged attributes we can't assume that len == reply->length.  
					  * Use reply->length for copying the string data into the packet.  Use len for the
					 *  true length of the string+tags.
					 */
					memcpy(ptr, reply->strvalue, reply->length);
					ptr += reply->length;
					total_length += len;
					break;
				
				case PW_TYPE_INTEGER:
				case PW_TYPE_IPADDR:
					*length_ptr += 4;
					if (vsa_length_ptr) *vsa_length_ptr += 4;

					if (reply->type == PW_TYPE_INTEGER ) {
				        	/*  If tagged, the tag becomes the MSB of the value */
				        	if(reply->flags.has_tag) {
							/*  Tag must be ( 0x01 -> 0x1F ) OR 0x00  */
							if(!TAG_VALID(reply->flags.tag)) {
								reply->flags.tag = 0x00;
							}
							lvalue = htonl((reply->lvalue & 0xffffff) | ((reply->flags.tag & 0xff) << 24));
						} else {
							lvalue = htonl(reply->lvalue);
						}
					} else {
						/*
						 *  IP address is already in network byte order.
						 */
						lvalue = reply->lvalue;
					}
					memcpy(ptr, &lvalue, 4);
					ptr += 4;
					total_length += 4;
					break;

				/*
				 *  There are no tagged date attributes.
				 */
				case PW_TYPE_DATE:
					*length_ptr += 4;
					if (vsa_length_ptr) *vsa_length_ptr += 4;
					lvalue = htonl(reply->lvalue);
					memcpy(ptr, &lvalue, 4);
					ptr += 4;
					total_length += 4;
					break;
				default:
					break;
			}
			
		} /* done looping over all attributes */

		/*
		 *	Fill in the rest of the fields, and copy
		 *	the data over from the local stack to
		 *	the newly allocated memory.
		 *
		 *	Yes, all this 'memcpy' is slow, but it means
		 *	that we only allocate the minimum amount of
		 *	memory for a request.
		 */
		packet->data_len = total_length;
		packet->data = (uint8_t *) malloc(packet->data_len);
		if (!packet->data) {
			ast_log(LOG_ERROR, "rad_send: Out of memory");
			return -1;
		}
		memcpy(packet->data, data, packet->data_len);
		hdr = (radius_packet_t *) packet->data;

		total_length = htons(total_length);
		memcpy(hdr->length, &total_length, sizeof(u_short));

		/*
		 *	If this is not an authentication request, we
		 *	need to calculate the md5 hash over the entire packet
		 *	and put it in the vector.
		 */
		secretlen = strlen(secret);
		if (packet->code != PW_AUTHENTICATION_REQUEST && packet->code != PW_STATUS_SERVER) {
			struct MD5Context	context;
			/*
			 *	Set the Message-Authenticator attribute,
			 *	BEFORE setting the reply authentication vector
			 *	for CHALLENGE, ACCEPT and REJECT.
			 */
			if (msg_auth_offset) {
				uint8_t calc_auth_vector[AUTH_VECTOR_LEN];

				switch (packet->code) {
					default:
						break;
				
					case PW_AUTHENTICATION_ACK:
					case PW_AUTHENTICATION_REJECT:
					case PW_ACCESS_CHALLENGE:
						if (original) {
							memcpy(hdr->vector, original->vector, AUTH_VECTOR_LEN);
						}
						break;
				}

				memset(packet->data + msg_auth_offset + 2, 0, AUTH_VECTOR_LEN);
				lrad_hmac_md5(packet->data, packet->data_len, secret, secretlen, calc_auth_vector);
				memcpy(packet->data + msg_auth_offset + 2, calc_auth_vector, AUTH_VECTOR_LEN);
				memcpy(hdr->vector, packet->vector, AUTH_VECTOR_LEN);
		    }

		    MD5Init(&context);
		    MD5Update(&context, packet->data, packet->data_len);
		    MD5Update(&context, secret, strlen(secret));
		    MD5Final(digest, &context);

		    memcpy(hdr->vector, digest, AUTH_VECTOR_LEN);
		    memcpy(packet->vector, digest, AUTH_VECTOR_LEN);
		} else if (msg_auth_offset) {
			/*
			 *	Set the Message-Authenticator attribute,
			 *	AFTER setting the authentication vector
			 *	only for ACCESS-REQUESTS
			 */
			
			uint8_t calc_auth_vector[AUTH_VECTOR_LEN];

			switch (packet->code) {
				default:
					break;
				case PW_AUTHENTICATION_ACK:
				case PW_AUTHENTICATION_REJECT:
				case PW_ACCESS_CHALLENGE:
					if (original) {
						memcpy(hdr->vector, original->vector, AUTH_VECTOR_LEN);
					}
					break;
			}

			memset(packet->data + msg_auth_offset + 2, 0, AUTH_VECTOR_LEN);
			lrad_hmac_md5(packet->data, packet->data_len, secret, secretlen, calc_auth_vector);
			memcpy(packet->data + msg_auth_offset + 2, calc_auth_vector, AUTH_VECTOR_LEN);
			memcpy(hdr->vector, packet->vector, AUTH_VECTOR_LEN);
		}

		/*
		 *	If packet->data points to data, then we print out
		 *	the VP list again only for debugging.
		 */
	} 
	
	/*
	 *	And send it on it's way.
	 */
	sa = (struct sockaddr_in *) &saremote;
        memset ((char *) sa, '\0', sizeof (saremote));
	sa->sin_family = AF_INET;
	sa->sin_addr.s_addr = packet->dst_ipaddr;
	sa->sin_port = htons(packet->dst_port);

	return sendto(packet->sockfd, packet->data, (int)packet->data_len, 0, (struct sockaddr *)&saremote, sizeof(struct sockaddr_in));
}

int dict_vendorname(const char *name)
{
	DICT_VENDOR *v;

	/*
	 *	Find the vendor, if any.
	 */
	v = dictionary_vendors;
	while (v) {
		if (!strcasecmp(name, v->vendorname)) {
			return v->vendorpec;
		}
		v = v->next;
	}

	return 0;
}

int dict_addvendor(const char *name, int value)
{
	DICT_VENDOR *vval;

	if (value >= (1 << 16)) {
	     	ast_log(LOG_ERROR, "dict_addvendor: Cannot handle vendor ID larger than 65535");
		return -1;
	} else if (strlen(name) > (sizeof(vval->vendorname) -1)) {
		ast_log(LOG_ERROR, "dict_addvendor: vendor name too long");
		return -1;
	}

	if ((vval =(DICT_VENDOR *)malloc(sizeof(DICT_VENDOR))) == NULL) {
		ast_log(LOG_ERROR,"dict_addvendor: out of memory");
		return -1;
	}
	strncpy(vval->vendorname, name, sizeof(vval->vendorname)-1);
	vval->vendorpec  = value;

	/* Insert at front. */
	vval->next = dictionary_vendors;
	dictionary_vendors = vval;

	return 0;
}

int dict_addattr(const char *name, int vendor, int type, int value, ATTR_FLAGS flags)
{
	static int      max_attr = 0;
	DICT_ATTR	*attr;

	if (strlen(name) > (sizeof(attr->name) -1)) {
		ast_log(LOG_ERROR, "dict_addattr: attribute name too long");
		return -1;
	}

	/*
	 *	If the value is '-1', that means use a pre-existing
	 *	one (if it already exists).  If one does NOT already exist,
	 *	then create a new attribute, with a non-conflicting value,
	 *	and use that.
	 */
	if (value == -1) {
		attr = dict_attrbyname(name);
		if (attr != NULL) {
			return 0; /* exists, don't add it again */
		}

		value = ++max_attr;

	} else if (vendor == 0) {
		/*
		 *  Update 'max_attr'
		 */
		if (value > max_attr) 
			max_attr = value;
	}


	/*
	 *	Create a new attribute for the list
	 */
	if ((attr = (DICT_ATTR *)malloc(sizeof(DICT_ATTR))) == NULL) {
		ast_log(LOG_ERROR, "dict_addattr: out of memory");
		return -1;
	}
	strncpy(attr->name, name, sizeof(attr->name)-1);
	attr->attr = value;
	attr->type = type;
	attr->flags = flags;

	if (vendor) {
		attr->attr |= (vendor << 16);
	} else if ((attr->attr >= 0) && (attr->attr < 256)) {
		/*
		 *	If it's an on-the-wire base attribute,
		 *	then keep a quick reference to it, for speed.
		 */
		base_attributes[attr->attr] = attr;
	}

	/*
	 *	Add to the front of the list, so that
	 *	values at the end of the file override
	 *	those in the begin.
	 */
	attr->next = dictionary_attributes;
	dictionary_attributes = attr;
	
	return 0;
}

int dict_addvalue(const char *namestr, char *attrstr, int value)
{
	DICT_VALUE	*dval;

	if (strlen(namestr) > (sizeof(dval->name) -1)) {
		ast_log(LOG_ERROR, "dict_addvalue: value name too long");
		return -1;
	}

	if (strlen(attrstr) > (sizeof(dval->attrname) -1)) {
		ast_log(LOG_ERROR,"dict_addvalue: attribute name too long");
		return -1;
	}

	if ((dval = (DICT_VALUE *)malloc(sizeof(DICT_VALUE))) == NULL) {
		ast_log(LOG_ERROR, "dict_addvalue: out of memory");
		return -1;
	}
	
	strncpy(dval->name, namestr, sizeof(dval->name)-1);
	strncpy(dval->attrname, attrstr,sizeof(dval->attrname)-1);
	dval->attr = 0;		/* ??? */
	dval->value = value;
	
	/* Insert at front. */
	dval->next = dictionary_values;
	dictionary_values = dval;
			
	return 0;
}

DICT_VALUE * dict_valbyname(int attr, const char *name)
{
	DICT_VALUE	*v;

	for (v = dictionary_values; v; v = v->next) {
		if ((attr == 0 || v->attr == attr) &&
		    strcmp(v->name, name) == 0)
		 return v;
               
	}

	return NULL;
}

DICT_ATTR * dict_attrbyvalue(int val)
{
	DICT_ATTR	*a;

	/*
	 *	If it's an on-the-wire base attribute, return
	 *	the cached value for it.
	 */
	if ((val >= 0) && (val < 256)) {
		return base_attributes[val];
	}

	for (a = dictionary_attributes; a; a = a->next) {
		if (a->attr == val)
			return a;
	}

	return NULL;
}

DICT_ATTR * dict_attrbyname(const char *name)
{
	DICT_ATTR	*a;

	a = dictionary_attributes;
	while (a) {
		if (!strcasecmp(a->name, name)) 
			break;
		a = a->next;
	}

	if (a) {
		return a;
	} else {
		return NULL;
	}
}

static int my_dict_init(const char *dir, const char *fn, const char *src_file, int src_line)
{
	FILE	*fp;
	char 	dirtmp[256];
	char	buf[256];
	char	namestr[256];
	char	valstr[256];
	char	attrstr[256];
	char	typestr[256];
	char	optstr[256];
	char	*p, *s, *c;
	char	*keyword;
	char	*data;
	int	line = 0;
	int	value;
	int	type;
	int	vendor;
	int	block_vendor;
	int	is_attrib;
	int	vendor_usr_seen = 0;
	int	is_nmc;
	ATTR_FLAGS  flags;

	if (strlen(fn) >= sizeof(dirtmp) / 2 || strlen(dir) >= sizeof(dirtmp) / 2) {
		ast_log(LOG_ERROR,"dict_init: filename name too long");
		return -1;
	}

	/*
	 *	First see if fn is relative to dir. If so, create
	 *	new filename. If not, remember the absolute dir.
	 */
	p = strrchr(fn, '/');
	if (p) {
		strcpy(dirtmp, fn);
		dirtmp[p - fn] = 0;
		dir = dirtmp;
	} else if (dir && dir[0] && strcmp(dir, ".") != 0) {
		snprintf(dirtmp, sizeof(dirtmp), "%s/%s", dir, fn);
		fn = dirtmp;
	}

	fp = fopen(fn, "r");
	if (!fp) {
		if (!src_file) {
			ast_log(LOG_ERROR,"Couldn't open dictionary \"%s\": %s\n",fn, strerror(errno));
		} else {
			ast_log(LOG_ERROR,"%s[%d]: Couldn't open dictionary \"%s\": %s\n", src_file, src_line, fn, strerror(errno));
		}
		return -1;
	}

	block_vendor = 0;

	while (fgets(buf, sizeof(buf), fp) != NULL) {

		line++;
		if (buf[0] == '#' || buf[0] == 0 ||
		  buf[0] == '\n' || buf[0] == '\r')
			continue;

		/*
		 *  Comments should NOT be appearing anywhere but
		 *  as comments;
		 */
		p = strchr(buf, '#');
		if (p) *p = '\0';

		keyword = strtok(buf, " \t\r\n");
		if (!keyword)
			continue;

		data = strtok(NULL, "\r\n");
		if (!data || data[0] == 0) {
			ast_log(LOG_ERROR, "dict_init: %s[%d]: invalid entry", fn, line);
			fclose(fp);
			return -1;
		}

		/*
		 *	See if we need to import another dictionary.
		 */
		if (!strcasecmp(keyword, "$INCLUDE")) {
			p = strtok(data, " \t");
			if (my_dict_init(dir, data, fn, line) < 0) {
				fclose(fp);
				return -1;
			}
			continue;
		} /* $INCLUDE */

		/*
		 *	Perhaps this is an attribute.
		 */
		is_attrib = 0;
		is_nmc = 0;
		if (!strcasecmp(keyword, "ATTRIBUTE"))
			is_attrib = 1;
		if (!strcasecmp(keyword, "ATTRIB_NMC")) {
			is_attrib = 1;
			is_nmc = 1;
		}
		
		if (is_attrib) {
			vendor = 0;
			optstr[0] = 0;
			if(sscanf(data, "%s%s%s%s", namestr, valstr, typestr, optstr) < 3) {
				ast_log(LOG_ERROR,"dict_init: %s[%d]: invalid ATTRIBUTE line", fn, line);
				fclose(fp);
				return -1;
			}

			/*
			 *	Convert ATTRIB_NMC into our format.
			 *	We might need to add USR to the list of
			 *	vendors first.
			 */
			if (is_nmc && optstr[0] == 0) {
				if (!vendor_usr_seen) {
					if (dict_addvendor("USR", VENDORPEC_USR) < 0) {
						ast_log(LOG_ERROR,"dict_init: %s[%d]: %s", fn, line, librad_errstr);
						fclose(fp);
						return -1;
					}
					vendor_usr_seen = 1;
				}
				strcpy(optstr, "USR");
			}

			/*
			 *	Validate all entries
			 */
			if (!isdigit((int) *valstr)) {
				ast_log(LOG_ERROR,"dict_init: %s[%d]: invalid value", fn, line);
				fclose(fp);
				return -1;
			}
			if (valstr[0] != '0')
				value = atoi(valstr);
			else
				sscanf(valstr, "%i", &value);

			/*
			 *	find the type.
			 */
			for (type = 0; dtypes[type]; type++) {
				if (!strcasecmp(typestr, dtypes[type]))
					break;
			}
			if (!dtypes[type]) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: invalid type", fn, line);
				fclose(fp);
				return -1;
			}

			/*
			 *	Ignore comments
			 */
			if (optstr[0] == '#') {
				optstr[0] = '\0';
			}

			/*
			 *	Only look up the vendor if the string
			 *	is non-empty.
			 */

			memset(&flags, 0, sizeof(flags));
			s = strtok(optstr, ",");
			while(s) {
			      if (!strcmp(s, "has_tag") || !strcmp(s, "has_tag=1")) {
				       /* Boolean flag, means this is a
					  tagged attribute */
				       flags.has_tag = 1;
				} else if (!strncmp(s, "len+=", 5) || !strncmp(s, "len-=", 5)) { 
				        /* Length difference, to accomodate
					   braindead NASes & their vendors */
				        flags.len_disp = strtol(s + 5, &c, 0);
					if (*c) {
					      ast_log(LOG_ERROR, "dict_init: %s[%d] invalid option %s", fn, line, s);
						fclose(fp);
						return -1;
					}
					if (s[3] == '-') {
					      flags.len_disp = -flags.len_disp;
					}
				} else if (!strncmp(s, "encrypt=", 8)) {
				        /* Encryption method, defaults to 0 (none).
					   Currently valid is just type 2,
					   Tunnel-Password style, which can only
					   be applied to strings. */    
				        flags.encrypt = strtol(s + 8, &c, 0);
					if (*c) {
					      ast_log(LOG_ERROR, "dict_init: %s[%d] invalid option %s", fn, line, s);
						fclose(fp);
						return -1;
					}
				} else {
				        /* Must be a vendor 'flag'... */
				        if (strncmp(s, "vendor=", 7) == 0) {
					      /* New format */
					      s += 7;   
					}
					vendor = dict_vendorname(s);
					if (!vendor) {
					      ast_log(LOG_ERROR, "dict_init: %s[%d]: unknown vendor %s", fn, line, optstr);
						fclose(fp);
						return -1;
					}
					if (block_vendor && optstr[0] && (block_vendor != vendor)) {
					      ast_log(LOG_ERROR, "dict_init: %s[%d]: mismatched vendor %s within BEGIN-VENDOR/END-VENDOR block", fn, line, optstr);
						fclose(fp);
						return -1;
					}
				}
				s = strtok(NULL, ",");
			}
 
			if (block_vendor) vendor = block_vendor;

			if (dict_addattr(namestr, vendor, type, value, flags) < 0) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: %s", fn, line, librad_errstr);
				fclose(fp);
				return -1;
			}
			continue;
		} /* ATTRIBUTE, or ATTRIB_NMC */

		/*
		 *	Process VALUE lines.
		 */
		if (!strcasecmp(keyword, "VALUE")) {
			if (sscanf(data, "%s%s%s", attrstr, namestr, valstr) != 3) {
				ast_log(LOG_ERROR,"dict_init: %s[%d]: invalid VALUE line", fn, line);
				fclose(fp);
				return -1;
			}
			/*
			 *	For Compatibility, skip "Server-Config"
			 */
			if (!strcasecmp(attrstr, "Server-Config"))
				continue;

			/*
			 *	Validate all entries
			 */
			if (!isdigit((int) *valstr)) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: invalid value", fn, line);
				fclose(fp);
				return -1;
			}
			if (valstr[0] != '0')
				value = atoi(valstr);
			else
				sscanf(valstr, "%i", &value);

			if (dict_addvalue(namestr, attrstr, value) < 0) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: %s", fn, line, librad_errstr);
				fclose(fp);
				return -1;
			}
			continue;
		} /* VALUE */

		/*
		 *	Process VENDOR lines.
		 */
		if (!strcasecmp(keyword, "VENDOR")) {
			if (sscanf(data, "%s%s", attrstr, valstr) != 2) {
				ast_log(LOG_ERROR, "dict_init: %s[%d] invalid VENDOR entry", fn, line);
				fclose(fp);
				return -1;
			}

			/*
			 *	 Validate all entries
			 */
			if (!isdigit((int) *valstr)) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: invalid value", fn, line);
				fclose(fp);
				return -1;
			}
			value = atoi(valstr);

			/* Create a new VENDOR entry for the list */
			if (dict_addvendor(attrstr, value) < 0) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: %s", fn, line, librad_errstr);
				fclose(fp);
				return -1;
			}

			if (value == VENDORPEC_USR)
				vendor_usr_seen = 1;

			continue;
		} /* VENDOR */

		if (!strcasecmp(keyword, "BEGIN-VENDOR")) {
			optstr[0] = 0;
			if (sscanf(data, "%s", optstr) != 1) {
				ast_log(LOG_ERROR, "dict_init: %s[%d] invalid BEGIN-VENDOR entry", fn, line);
				fclose(fp);
				return -1;
			}

			vendor = dict_vendorname(optstr);
			if (!vendor) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: unknown vendor %s", fn, line, optstr);
				fclose(fp);
				return -1;
			}
			block_vendor = vendor;
			continue;
		} /* BEGIN-VENDOR */

		if (!strcasecmp(keyword, "END-VENDOR")) {
			optstr[0] = 0;
			if (sscanf(data, "%s", optstr) != 1) {
				ast_log(LOG_ERROR, "dict_init: %s[%d] invalid END-VENDOR entry", fn, line);
				fclose(fp);
				return -1;
			}

			vendor = dict_vendorname(optstr);
			if (!vendor) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: unknown vendor %s", fn, line, optstr);
				fclose(fp);
				return -1;
			}

			if (vendor != block_vendor) {
				ast_log(LOG_ERROR, "dict_init: %s[%d]: END-VENDOR %s does not match any previous BEGIN-VENDOR", fn, line, optstr);
				fclose(fp);
				return -1;
			}
			block_vendor = 0;
			continue;
		} /* END-VENDOR */

		/*
		 *	Any other string: We don't recognize it.
		 */
		ast_log(LOG_ERROR, "dict_init: %s[%d] invalid keyword \"%s\"", fn, line, keyword);
		fclose(fp);
		return -1;
	}
	fclose(fp);
	return 0;
}

static void dict_free(void)
{
	DICT_ATTR	*dattr, *anext;
	DICT_VALUE	*dval, *vnext;
	DICT_VENDOR	*dvend, *enext;

	dattr = dictionary_attributes;
	while (dattr) {
		anext = dattr->next;
		free(dattr);
		dattr = anext;
	}
	
	dval = dictionary_values;
	while (dval) {
		vnext = dval->next;
		free(dval);
		dval = vnext;
	}
	
	dvend = dictionary_vendors;
	while (dvend) {
		enext = dvend->next;
		free(dvend);
		dvend = enext;
	}

	dictionary_attributes = NULL;
	dictionary_values = NULL;
	dictionary_vendors = NULL;

	memset(base_attributes, 0, sizeof(base_attributes));
}

int dict_init(const char *dir, const char *fn)
{
	DICT_ATTR	*attr;
	DICT_VALUE	*dval;

	dict_free();
	if (my_dict_init(dir, fn, NULL, 0) < 0)
		return -1;
	
	dval = dictionary_values;
	while (dval) {
		if (dval->attr == 0) {
			attr = dict_attrbyname(dval->attrname);
			if (attr == NULL) {
				ast_log(LOG_ERROR,"dict_init: VALUE %s for unknown ATTRIBUTE %s",dval->name, dval->attrname);
				return -1;
			}
			dval->attr = attr->attr;
		}
		dval = dval->next;
	}
	return 0;
}

static const char *vp_tokens[] = {
  "?",				/* T_INVALID */
  "EOL",			/* T_EOL */
  "{",
  "}",
  "(",
  ")",
  ",",
  ";",
  "+=",
  "-=",
  ":=",
  "=",
  "!=",
  ">=",
  ">",
  "<=",
  "<",
  "=~",
  "!~",
  "=*",
  "~*",
  "==",
  "#",
  "<BARE-WORD>",
  "<\"STRING\">",
  "<'STRING'>",
  "<`STRING`>"
};

void librad_safeprint(char *in, int inlen, char *out, int outlen)
{
	unsigned char	*str = (unsigned char *)in;
	int		done = 0;
	int		sp = 0;

	if (inlen < 0) inlen = strlen(in);

	while (inlen-- > 0 && (done + 3) < outlen) {
		/*
		 *	Hack: never print trailing zero.
		 *	Some clients send strings with an off-by-one
		 *	length (confused with strings in C).
		 */
		if (inlen == 0 && *str == 0)
			break;

		sp = 0;

		switch (*str) {
			case '\\':
				sp = '\\';
				break;
			case '\r':
				sp = 'r';
				break;
			case '\n':
				sp = 'n';
				break;
			case '\t':
				sp = 't';
				break;
			default:
				if (*str < 32 || (*str >= 128)){
					snprintf(out, outlen, "\\%03o", *str);
					done += 4;
					out  += 4;
					outlen -= 4;
				} else {
					*out++ = *str;
					outlen--;
					done++;
				}
		}
		if (sp) {
			*out++ = '\\';
			*out++ = sp;
			outlen -= 2;
			done += 2;
		}
		str++;
	}
	*out = 0;
}

int vp_prints_value(char * out, int outlen, VALUE_PAIR *vp, int delimitst)
{
	DICT_VALUE  *v;
	char        buf[1024];
	char        *a;
	time_t      t;
	struct tm   s_tm;

	out[0] = 0;
	if (!vp) return 0;

	switch (vp->type) {
		case PW_TYPE_STRING:
			if (vp->attribute == PW_NAS_PORT_ID)
				a = (char *)vp->strvalue;
			else {
				if (delimitst && vp->flags.has_tag) {
				        /* Tagged attribute: print delimter and ignore tag */
				        buf[0] = '"';
					librad_safeprint((char *)(vp->strvalue),
							 vp->length, buf + 1, sizeof(buf) - 2);
					strcat(buf, "\"");
				} else if (delimitst) {
				        /* Non-tagged attribute: print delimter */
				        buf[0] = '"';
					librad_safeprint((char *)vp->strvalue,
							 vp->length, buf + 1, sizeof(buf) - 2);
					strcat(buf, "\"");
				} else {
				        /* Non-tagged attribute: no delimiter */
				        librad_safeprint((char *)vp->strvalue,
							 vp->length, buf, sizeof(buf));
				}
				a = buf;
			}
			break;
		case PW_TYPE_INTEGER:
		        if ( vp->flags.has_tag ) {
			        /* Attribute value has a tag, need to ignore it */
			        if ((v = dict_valbyattr(vp->attribute, (vp->lvalue & 0xffffff)))
				    != NULL)
				        a = v->name;
				else {
				        snprintf(buf, sizeof(buf), "%u", (vp->lvalue & 0xffffff));
				        a = buf;
				}
			} else {
			        /* Normal, non-tagged attribute */
			        if ((v = dict_valbyattr(vp->attribute, vp->lvalue))
				    != NULL)
				        a = v->name;
				else {
				        snprintf(buf, sizeof(buf), "%u", vp->lvalue);
					a = buf;
				}
			}
			break;
		case PW_TYPE_DATE:
			t = vp->lvalue;
			if (delimitst) {
			  strftime(buf, sizeof(buf), "\"%b %e %Y\"", gmtime_r(&t, &s_tm));
			} else {
			  strftime(buf, sizeof(buf), "%b %e %Y", gmtime_r(&t, &s_tm));
			}
			a = buf;
			break;
		case PW_TYPE_IPADDR:
			if (vp->strvalue[0])
				a = (char *)vp->strvalue;
			else
				a = ip_hostname((char *)vp->strvalue,
						sizeof(vp->strvalue),
						vp->lvalue);
			break;
		case PW_TYPE_ABINARY:
#ifdef ASCEND_BINARY
		  a = buf;
		  print_abinary(vp, (unsigned char *)buf, sizeof(buf));
		  break;
#else
		  /* FALL THROUGH */
#endif
		case PW_TYPE_OCTETS:
		  strcpy(buf, "0x");
		  a = buf + 2;
		  for (t = 0; t < vp->length; t++) {
			sprintf(a, "%02x", vp->strvalue[t]);
			a += 2;
		  }
		  a = buf;
		  break;

		default:
			a = 0;
			break;
	}
	strncpy(out, a?a:"UNKNOWN-TYPE", outlen);
	
	return strlen(out);
}

int vp_prints(char *out, int outlen, VALUE_PAIR *vp)
{
	int		len;
	const char	*token = NULL;

	out[0] = 0;
	if (!vp) return 0;

	if (strlen(vp->name) + 3 > (size_t)outlen) {
		return 0;
	}

	if ((vp->operator > T_INVALID) &&
	    (vp->operator < T_TOKEN_LAST)) {
		token = vp_tokens[vp->operator];
	} else {
		token = "<INVALID-TOKEN>";
	}

	if( vp->flags.has_tag ) {

		snprintf(out, outlen, "%s:%d %s ", vp->name, vp->flags.tag,
			 token);
		
		len = strlen(out);
		vp_prints_value(out + len, outlen - len, vp, 1);

	} else {

	        snprintf(out, outlen, "%s %s ", vp->name, token);
		len = strlen(out);
		vp_prints_value(out + len, outlen - len, vp, 1);

	}	  

	return strlen(out);
}

void vp_print(FILE *fp, VALUE_PAIR *vp)
{
	char	buf[1024];

	vp_prints(buf, sizeof(buf), vp);
	fputs(buf, fp);
}

void vp_printlist(FILE *fp, VALUE_PAIR *vp)
{
	for (; vp; vp = vp->next) {
		fprintf(fp, "\t");
		vp_print(fp, vp);
		fprintf(fp, "\n");
	}
}
